from functools import wraps
from bson.objectid import ObjectId, InvalidId
from .errors import ObjectSaveRequired


def resolve_id(id_):
    # ObjectId(None) apparently generates a new unique object id
    # which is not a behaviour we need
    if id_ is not None:
        try:
            objid_expr = ObjectId(id_)
            if str(objid_expr) == id_:
                id_ = objid_expr
        except (InvalidId, TypeError):
            pass
    return id_


def save_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        this = args[0]
        if this.is_new:
            raise ObjectSaveRequired("This object must be saved first")
        return func(*args, **kwargs)

    return wrapper


def pick_rw_shard():
    from .db import db
    from random import choice

    return choice(db.rw_shards)




class BaseModelHook(object):
    @classmethod
    def on_hook_register(cls, model_class):
        pass

    @classmethod
    def on_hook_unregister(cls, model_class):
        pass

    @classmethod
    def on_model_init(cls, model):
        pass

    def on_model_destroy(self, model):
        pass

    def on_model_save(self, model, is_new):
        pass



from unittest import TestCase
from ..uorm.db import db
from ..context import ctx, ContextError
from ..queue import DummyQueue


class MongoMockTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        db.configure(
            {
                "meta": {"uri": "mongodb://fake_server:27017/unittest_meta"},
                "shards": {
                    "s1": {"uri": "mongodb://fake_server:27017/unittest_s1"},
                    "s2": {"uri": "mongodb://fake_server:27017/unittest_s2"},
                    "s3": {"uri": "mongodb://fake_server:27017/unittest_s3"},
                    "s4": {"uri": "mongodb://fake_server:27017/unittest_s4"},
                },
            },
            mock=True,
        )
        try:
            _ = ctx.queue
            del ctx.queue
        except ContextError:
            pass
        ctx.queue = DummyQueue({})




import logging
from copy import deepcopy
from pymongo.errors import OperationFailure

from glasskit import ctx
from glasskit.uorm.errors import DoNotSave
from ..db import db
from .fields import ObjectIdField
from .meta import MetaModel
from .hooks import BaseModelHook


class BaseModel(MetaModel):

    __hooks_classes__ = None

    USE_INITIAL_STATE = False

    _id: ObjectIdField()

    def __init__(self, attrs=None, **kwargs):
        if attrs is None:
            attrs = {}

        for field in self.__fields__:
            value = attrs.get(field, self.__def_values__.get(field, None))
            if callable(value):
                value = value()
            self.__dict__[field] = value
        self._initial_state = None
        self.__set_initial_state()
        self.__hooks__ = []
        if self.__hooks_classes__:
            for hook_class in self.__hooks_classes__:
                hook_inst = hook_class.on_model_init(self)
                if hook_inst:
                    self.__hooks__.append(hook_inst)

    def validate(self):
        for name, func in self.__validators__.items():
            value = self.__dict__[name]
            func(value)

    @property
    def is_new(self):
        return self._id is None

    def __set_initial_state(self):
        if self.USE_INITIAL_STATE:
            self.__dict__["_initial_state"] = self.setup_initial_state()

    def setup_initial_state(self):
        # this method is highly ineffective and is higlhy recommended
        # to be overriden
        return deepcopy(self.to_dict(self.__fields__))

    @classmethod
    def register_model_hook(cls, model_hook_class, *args, **kwargs):
        if not issubclass(model_hook_class, BaseModelHook):
            raise TypeError("Invalid hook class")
        if cls.__hooks_classes__ is None:
            cls.__hooks_classes__ = set()
        if model_hook_class not in cls.__hooks_classes__:
            cls.__hooks_classes__.add(model_hook_class)
            model_hook_class.on_hook_register(cls, *args, **kwargs)
            ctx.log.debug(
                "Registered hook %s for model %s",
                model_hook_class.__name__,
                cls.__name__,
            )

    @classmethod
    def unregister_model_hook(cls, model_hook_class):
        if cls.__hooks_classes__ is None:
            return
        if model_hook_class in cls.__hooks_classes__:
            cls.__hooks_classes__.remove(model_hook_class)
            model_hook_class.on_hook_unregister(cls)

    @classmethod
    def clear_hooks(cls):
        if cls.__hooks_classes__ is None:
            return
        for hook_class in cls.__hooks_classes__.copy():
            cls.unregister_model_hook(hook_class)

    def _before_save(self):
        pass

    def _before_validation(self):
        pass

    def _before_delete(self):
        pass

    def _after_save(self, is_new):
        pass

    def _after_delete(self):
        pass

    def _save_to_db(self):
        pass

    def _delete_from_db(self):
        pass

    def invalidate(self, **kwargs):
        pass

    def destroy(self, skip_callback=False, invalidate_cache=True):
        if self.is_new:
            return
        if not skip_callback:
            self._before_delete()

        if invalidate_cache:
            self.invalidate()

        self._delete_from_db()
        if not skip_callback:
            self._after_delete()

        self._id = None
        for hook in self.__hooks__:
            try:
                hook.on_model_destroy(self)
            except Exception as e:
                ctx.log.error(
                    "error executing destroy hook %s on model %s(%s): %s",
                    hook.__class__.__name__,
                    self.__class__.__name__,
                    self._id,
                    e,
                )
        return self

    def save(self, skip_callback=False, invalidate_cache=True):
        is_new = self.is_new

        if not skip_callback:
            try:
                self._before_validation()
            except DoNotSave:
                return
        self.validate()

        if not skip_callback:
            try:
                self._before_save()
            except DoNotSave:
                return

        if invalidate_cache:
            self.invalidate()

        self._save_to_db()

        for hook in self.__hooks__:
            try:
                hook.on_model_save(self, is_new)
            except Exception as e:
                ctx.log.error(
                    "error executing save hook %s on model %s(%s): %s",
                    hook.__class__.__name__,
                    self.__class__.__name__,
                    self._id,
                    e,
                )

        self.__set_initial_state()
        if not skip_callback:
            self._after_save(is_new)

        return self

    def __repr__(self):
        attributes = ["%s=%r" % (a, getattr(self, a)) for a in self.__fields__]
        return "%s(\n    %s\n)" % (self.__class__.__name__, ",\n    ".join(attributes))

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        for field in self.__fields__:
            if hasattr(self, field):
                if not hasattr(other, field):
                    return False
                if getattr(self, field) != getattr(other, field):
                    return False
            elif hasattr(other, field):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self, fields=None, include_restricted=False):
        if fields is None:
            fields = self.__fields__

        result = {}
        for field in fields:
            if field in self.__restricted_fields__ and not include_restricted:
                continue
            if not hasattr(self, field):
                continue
            value = getattr(self, field)
            if callable(value):
                continue
            result[field] = value
        return result

    @classmethod
    def create(cls, **attrs):
        return cls(attrs)

    @classmethod
    def _ctor(cls, attrs, **kwargs):
        return cls(attrs, **kwargs)

    @classmethod
    def __get_possible_databases(cls):
        return [db.meta]

    @classmethod
    def ensure_indexes(cls, loud=False, overwrite=False):
        for keys, options in cls.__def_indexes__:
            for db in cls.__get_possible_databases():
                try:
                    db.conn[cls.collection].create_index(keys, **options)
                except OperationFailure as e:
                    if (
                        e.details.get("codeName") == "IndexOptionsConflict"
                        or e.details.get("code") == 85
                    ):
                        if overwrite:
                            if loud:
                                ctx.log.debug("Dropping index %s as conflicting", keys)
                            db.conn[cls.collection].drop_index(keys)
                            if loud:
                                ctx.log.debug(
                                    "Creating index with options: %s, %s", keys, options
                                )
                            db.conn[cls.collection].create_index(keys, **options)
                        else:
                            ctx.log.error(
                                "Index %s conflicts with an existing one, use overwrite param to fix it",
                                keys,
                            )

    def _reload_from_obj(self, obj):
        for field in self.__fields__:
            if field == "_id":
                continue
            value = getattr(obj, field)
            self.__dict__[field] = value



from flask.json import JSONEncoder
from bson import ObjectId, Timestamp
from .uorm.models.base_model import BaseModel
from .uorm.db import ObjectsCursor


class MongoJSONEncoder(JSONEncoder):
    def default(self, o):  # pylint: disable=method-hidden

        if isinstance(o, ObjectId):
            return str(o)
        if isinstance(o, Timestamp):
            return o.time
        if isinstance(o, (ObjectsCursor, set)):
            return list(o)
        if isinstance(o, BaseModel):
            return o.to_dict()

        return JSONEncoder.default(self, o)



from . import Command
from glasskit.queue.worker import BaseWorker
import importlib
import inspect


class Tasks(Command):
    def find_worker_class(self):
        wrk_module = importlib.import_module(f"{self.app.app_name}.tasks.worker")
        for objname in dir(wrk_module):
            obj = getattr(wrk_module, objname)
            if (
                inspect.isclass(obj)
                and issubclass(obj, BaseWorker)
                and obj is not BaseWorker
            ):
                return obj
        return None

    def run(self):
        wrk_cls = self.find_worker_class()
        if wrk_cls is None:
            raise RuntimeError(
                f"Worker class not found in {self.app.app_name}.tasks.worker"
            )
        wrk = wrk_cls()
        wrk.process_tasks()




import logging
from time import sleep
from datetime import datetime
from bson.objectid import ObjectId, InvalidId
from pymongo import MongoClient, ReturnDocument
from mongomock import MongoClient as MockMongoClient
from pymongo.errors import ServerSelectionTimeoutError
from pymongo.uri_parser import parse_uri
from contextlib import contextmanager
from glasskit.uorm.errors import AbortTransaction, InvalidShardId, ConfigurationError
from glasskit import ctx

MONGO_RETRY_MAX = 6
MONGO_RETRY_SLEEP = 3  # 3 seconds


class NoCache:

    NAME = "NoCache"

    @staticmethod
    def has(_key):
        return False

    @staticmethod
    def set(*_args, **_kwargs):
        return False

    @staticmethod
    def get(_key):
        return None

    @staticmethod
    def delete(_key):
        return False


def intercept_pymongo_errors(func):
    def wrapper(*args, **kwargs):
        retries = MONGO_RETRY_MAX
        while True:
            retries -= 1
            try:
                return func(*args, **kwargs)
            except ServerSelectionTimeoutError:
                ctx.log.error(
                    "ServerSelectionTimeoutError in db module, retries left = %d",
                    retries,
                )
                if retries == MONGO_RETRY_MAX // 2:
                    ctx.log.error("trying to reinstantiate connection")
                    db_inst = args[0]  # self
                    db_inst.reset_conn()
                elif retries == 0:
                    ctx.log.error(
                        "ServerSelectionTimeoutError max retries exceeded, giving up"
                    )
                    raise
                sleep(MONGO_RETRY_SLEEP)

    return wrapper


class ObjectsCursor:
    def __init__(self, cursor, ctor, shard_id=None):
        self.ctor = ctor
        self.cursor = cursor
        self._shard_id = shard_id

    def all(self):
        return list(self)

    def limit(self, *args, **kwargs):
        self.cursor.limit(*args, **kwargs)
        return self

    def skip(self, *args, **kwargs):
        self.cursor.skip(*args, **kwargs)
        return self

    def sort(self, *args, **kwargs):
        self.cursor.sort(*args, **kwargs)
        return self

    def __iter__(self):
        for item in self.cursor:
            if self._shard_id:
                yield self.ctor(shard_id=self._shard_id, attrs=item)
            else:
                yield self.ctor(attrs=item)

    def __getitem__(self, item):
        attrs = self.cursor.__getitem__(item)
        if self._shard_id:
            return self.ctor(shard_id=self._shard_id, attrs=attrs)
        return self.ctor(attrs)

    def __getattr__(self, item):
        return getattr(self.cursor, item)


class Shard:
    @contextmanager
    def transaction(self):
        client = self.get_client()
        try:
            self._session = client.start_session()
            self._session.start_transaction()
            yield self._session
            self._session.commit_transaction()
        except Exception as e:
            self._session.abort_transaction()
            if not isinstance(e, AbortTransaction):
                raise
        finally:
            self._session.end_session()
            self._session = None

    def __init__(self, dbconf, shard_id=None, mock=False):
        self._config = dbconf
        self._client = None
        self._conn = None
        self._ro_conn = None
        self._shard_id = shard_id
        self._session = None
        self._mock = mock
        self.readonly = self._config.get("readonly", False)

    @property
    def _dbname(self):
        return parse_uri(self._config["uri"]).get("database")

    def get_client(self):
        if not self._client:
            client_kwargs = self._config.get("pymongo_extra", {})
            if self._mock:
                self._client = MockMongoClient(self._config["uri"], **client_kwargs)
            else:
                self._client = MongoClient(self._config["uri"], **client_kwargs)
        return self._client

    def reset_conn(self):
        self._client = None
        self._conn = None

    def init_conn(self):
        ctx.log.info("creating a new mongo connection")
        client = self.get_client()
        self._conn = client[self._dbname]

    @property
    def conn(self):
        if self._conn is None:
            self.init_conn()
        return self._conn

    @intercept_pymongo_errors
    def get_obj(self, ctor, collection, query):
        if not isinstance(query, dict):
            try:
                query = {"_id": ObjectId(query)}
            except InvalidId:
                pass
        data = self.conn[collection].find_one(query, session=self._session)
        if data:
            if self._shard_id:
                return ctor(shard_id=self._shard_id, attrs=data)
            else:
                return ctor(attrs=data)

        return None

    @intercept_pymongo_errors
    def get_obj_id(self, collection, query):
        return self.conn[collection].find_one(
            query, projection=(), session=self._session
        )["_id"]

    @intercept_pymongo_errors
    def get_objs(self, ctor, collection, query, **kwargs):
        if self._session:
            kwargs["session"] = self._session
        cursor = self.conn[collection].find(query, **kwargs)
        return ObjectsCursor(cursor, ctor, shard_id=self._shard_id)

    @intercept_pymongo_errors
    def get_objs_projected(self, collection, query, projection, **kwargs):
        if self._session:
            kwargs["session"] = self._session
        cursor = self.conn[collection].find(query, projection=projection, **kwargs)
        return cursor

    @intercept_pymongo_errors
    def get_aggregated(self, collection, pipeline, **kwargs):
        if self._session:
            kwargs["session"] = self._session
        cursor = self.conn[collection].aggregate(pipeline, **kwargs)
        return cursor

    @intercept_pymongo_errors
    def count_docs(self, collection, query, **kwargs):
        return self.conn[collection].count_documents(query, **kwargs)

    def get_objs_by_field_in(self, cls, collection, field, values, **kwargs):
        return self.get_objs(cls, collection, {field: {"$in": values,},}, **kwargs)

    @intercept_pymongo_errors
    def save_obj(self, obj):
        if obj.is_new:
            # object to_dict() method should always return all fields
            data = obj.to_dict(include_restricted=True)
            # although with the new object we shouldn't pass _id=null to mongo
            del data["_id"]
            inserted_id = (
                self.conn[obj.collection]
                .insert_one(data, session=self._session)
                .inserted_id
            )
            obj._id = inserted_id
        else:
            self.conn[obj.collection].replace_one(
                {"_id": obj._id},
                obj.to_dict(include_restricted=True),
                upsert=True,
                session=self._session,
            )

    @intercept_pymongo_errors
    def delete_obj(self, obj):
        if obj.is_new:
            return
        self.conn[obj.collection].delete_one({"_id": obj._id}, session=self._session)

    @intercept_pymongo_errors
    def find_and_update_obj(self, obj, update, when=None):
        query = {"_id": obj._id}
        if when:
            assert "_id" not in when
            query.update(when)

        new_data = self.conn[obj.collection].find_one_and_update(
            query, update, return_document=ReturnDocument.AFTER, session=self._session
        )
        if new_data and self._shard_id:
            new_data["shard_id"] = self._shard_id
        return new_data

    @intercept_pymongo_errors
    def delete_query(self, collection, query):
        return self.conn[collection].delete_many(query, session=self._session)

    @intercept_pymongo_errors
    def update_query(self, collection, query, update):
        return self.conn[collection].update_many(query, update, session=self._session)

    # SESSIONS
    @intercept_pymongo_errors
    def get_session(self, sid, collection="sessions"):
        return self.conn[collection].find_one({"sid": sid})

    @intercept_pymongo_errors
    def update_session(self, sid, data, expiration, collection="sessions"):
        self.conn[collection].update(
            {"sid": sid}, {"sid": sid, "data": data, "expiration": expiration}, True
        )

    @intercept_pymongo_errors
    def cleanup_sessions(self, collection="sessions"):
        return self.conn[collection].remove({"expiration": {"$lt": datetime.now()}})[
            "n"
        ]


class DB:

    MONGODB_INFO_FIELDS = (
        "allocator",
        "bits",
        "debug",
        "gitVersion",
        "javascriptEngine",
        "maxBsonObjectSize",
        "modules",
        "ok",
        "openssl",
        "storageEngines",
        "sysInfo",
        "version",
        "versionArray",
    )

    @staticmethod
    def parse_shard_cfg(shard_name, shardcfg, extra):
        if "uri" not in shardcfg:
            raise ConfigurationError(f"uri is missing for shard {shard_name}")
        local_extra = shardcfg.get("pymongo_extra", {})
        return {"uri": shardcfg["uri"], "pymongo_extra": {**extra, **local_extra}}

    def parse_cfg(self, dbconf):
        config = {"shards": {}}
        global_extra = dbconf.get("pymongo_extra", {})
        if "meta" not in dbconf:
            raise ConfigurationError("meta section is missing from config")
        config["meta"] = self.parse_shard_cfg("meta", dbconf["meta"], global_extra)
        if "shards" in dbconf:
            for shard_id, shardcfg in dbconf["shards"].items():
                config["shards"][shard_id] = self.parse_shard_cfg(
                    shard_id, shardcfg, global_extra
                )
        return config

    def configure(self, dbconf, mock=False):
        config = self.parse_cfg(dbconf)
        self.__meta = Shard(config["meta"], mock=mock)
        self.__shards = {}
        if "shards" in config:
            for shard_id, config in config["shards"].items():
                self.__shards[shard_id] = Shard(config, shard_id=shard_id, mock=mock)
        self.initialized = True

    def __init__(self):
        self.__meta = None
        self.__shards = {}
        self.initialized = False
        self.l1c = NoCache()
        self.l2c = NoCache()

    @property
    def rw_shards(self):
        return [
            shard_id for shard_id, shard in self.shards.items() if not shard.readonly
        ]

    @property
    def meta(self):
        if not self.initialized:
            raise RuntimeError("attempted to use uninitialized DB")
        return self.__meta

    @property
    def shards(self):
        if not self.initialized:
            raise RuntimeError("attempted to use uninitialized DB")
        return self.__shards

    def get_shard(self, shard_id):
        if shard_id not in self.shards:
            raise InvalidShardId(shard_id)
        return self.shards[shard_id]

    def mongodb_info(self):
        def sys_info(raw_info):
            return {k: v for k, v in raw_info.items() if k in self.MONGODB_INFO_FIELDS}

        return dict(
            meta=sys_info(self.meta.conn.client.server_info()),
            shards={
                shard_id: sys_info(self.shards[shard_id].conn.client.server_info())
                for shard_id in self.shards
            },
        )

    def setup_cache(self, cache, level):
        if level not in (1, 2):
            raise ValueError("cache level must be either 1 or 2")
        attrs = ["has", "get", "set", "delete"]
        if not all(hasattr(cache, attr) for attr in attrs):
            raise TypeError(
                "a cache instance must have set, get, delete" " and has methods defined"
            )
        if level == 1:
            self.l1c = cache
        else:
            self.l2c = cache


db = DB()



import logging
from functools import partial
from time import time
from .base_model import BaseModel
from ..db import db
from ..utils import save_required, resolve_id, ObjectId
from ..errors import ModelDestroyed
from glasskit.errors import NotFound
from glasskit import ctx


class StorableModel(BaseModel):
    @property
    def _db(self):
        return db.meta

    def _save_to_db(self):
        self._db.save_obj(self)

    def update(self, data, skip_callback=False, invalidate_cache=True):
        for field in self.__fields__:
            if (
                field in data
                and field not in self.__rejected_fields__
                and field != "_id"
            ):
                self.__dict__[field] = data[field]
        self.save(skip_callback=skip_callback, invalidate_cache=invalidate_cache)

    @save_required
    def db_update(self, update, when=None, reload=True, invalidate_cache=True):
        """
                :param update: MongoDB update query
                :param when: filter query. No update will happen if it does not match
                :param reload: Load the new stat into the object (Caution: if you do not do this
                               the next save() will overwrite updated fields)
                :return: True if the document was updated. Otherwise - False
                """
        new_data = self._db.find_and_update_obj(self, update, when)
        if invalidate_cache and new_data:
            self.invalidate()

        if reload and new_data:
            tmp = self.__class__(new_data)
            self._reload_from_obj(tmp)

        return bool(new_data)

    def _delete_from_db(self):
        self._db.delete_obj(self)

    def _refetch_from_db(self):
        return self.find_one({"_id": self._id})

    def reload(self):
        if self.is_new:
            return
        tmp = self._refetch_from_db()
        if tmp is None:
            raise ModelDestroyed("model has been deleted from db")
        self._reload_from_obj(tmp)

    @classmethod
    # E.g. override if you want model to always return a subset of documents in its collection
    def _preprocess_query(cls, query):
        return query

    @classmethod
    def find(cls, query=None, **kwargs):
        if not query:
            query = {}
        return db.meta.get_objs(
            cls._ctor, cls.collection, cls._preprocess_query(query), **kwargs
        )

    @classmethod
    def aggregate(cls, pipeline, query=None, **kwargs):
        if not query:
            query = {}
        pipeline = [{"$match": cls._preprocess_query(query)}] + pipeline
        return db.meta.get_aggregated(cls.collection, pipeline, **kwargs)

    @classmethod
    def find_projected(cls, query=None, projection=("_id",), **kwargs):
        if not query:
            query = {}
        return db.meta.get_objs_projected(
            cls.collection,
            cls._preprocess_query(query),
            projection=projection,
            **kwargs,
        )

    @classmethod
    def find_one(cls, query, **kwargs):
        return db.meta.get_obj(
            cls._ctor, cls.collection, cls._preprocess_query(query), **kwargs
        )

    @classmethod
    def get(cls, expression, raise_if_none=None):
        if expression is None:
            return None

        expression = resolve_id(expression)
        if isinstance(expression, ObjectId):
            query = {"_id": expression}
        else:
            expression = str(expression)
            query = {cls.KEY_FIELD: expression}
        res = cls.find_one(query)
        if res is None and raise_if_none is not None:
            if isinstance(raise_if_none, Exception):
                raise raise_if_none
            else:
                raise NotFound(f"{cls.__name__} not found")
        return res

    @classmethod
    def destroy_all(cls):
        db.meta.delete_query(cls.collection, cls._preprocess_query({}))

    @classmethod
    def destroy_many(cls, query):
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        db.meta.delete_query(cls.collection, cls._preprocess_query(query))

    @classmethod
    def update_many(cls, query, attrs):
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        db.meta.update_query(cls.collection, cls._preprocess_query(query), attrs)

    @classmethod
    def cache_get(cls, expression, raise_if_none=None):
        if expression is None:
            return None
        cache_key = f"{cls.collection}.{expression}"
        getter = partial(cls.get, expression, raise_if_none)
        return cls._cache_get(cache_key, getter)

    @classmethod
    def _cache_get(cls, cache_key, getter, ctor=None):
        t1 = time()
        if not ctor:
            ctor = cls

        if db.l1c.has(cache_key):
            data = db.l1c.get(cache_key)
            td = time() - t1
            ctx.log.debug(
                "%s L1 hit %s %.3f secs", db.l1c.__class__.__name__, cache_key, td
            )
            return ctor(data)

        if db.l2c.has(cache_key):
            data = db.l2c.get(cache_key)
            db.l1c.set(cache_key, data)
            td = time() - t1
            ctx.log.debug(
                "%s L2 hit %s %.3f secs", db.l2c.__class__.__name__, cache_key, td
            )
            return ctor(data)

        obj = getter()
        if obj:
            data = obj.to_dict()
            db.l2c.set(cache_key, data)
            db.l1c.set(cache_key, data)

        td = time() - t1
        ctx.log.debug("%s miss %s %.3f secs", db.l2c.__class__.__name__, cache_key, td)
        return obj

    def invalidate(self):
        if self.USE_INITIAL_STATE:
            for field in self.__cache_key_fields__:
                value = self._initial_state.get(field)
                if value:
                    self._invalidate(f"{self.collection}.{value}")

    @staticmethod
    def _invalidate(cache_key):
        if db.l1c.delete(cache_key):
            ctx.log.debug("%s delete %s", db.l1c.__class__.__name__, cache_key)
        if db.l2c.delete(cache_key):
            ctx.log.debug("%s delete %s", db.l2c.__class__.__name__, cache_key)



from flask import Blueprint, g
from inspect import isclass
from .errors import AuthenticationError


class Controller(Blueprint):

    auth_error = AuthenticationError

    def __init__(self, name, import_name, require_auth=True, **kwargs):
        super().__init__(name, import_name, **kwargs)

        if not isclass(self.auth_error) or not issubclass(self.auth_error, Exception):
            raise TypeError(
                f"{self.__class__.__name__}.auth_error must be a subclass of Exception"
            )

        self.before_request(self.__set_auth_data)
        if require_auth:
            self.before_request(self.check_auth)

    def get_current_user(self) -> object:
        return None

    def __set_auth_data(self):
        g.user = self.get_current_user()

    def check_auth(self):
        if g.user is None:
            raise self.auth_error()



from .task import BaseTask
from ..context import ctx
DEFAULT_RETRIES = 5
DEFAULT_ACK_TIMEOUT = 1


class BaseQueue:

    ACK_POSTFIX = ":ack"

    def __init__(self, qcfg):
        self.cfg = qcfg

    def enqueue(self, task):
        if not isinstance(task, BaseTask):
            raise TypeError("only instances of Task are allowed")
        if task.received:
            raise RuntimeError("task is already received by a subscriber")
        ctx.log.debug("%s: enqueue task %s", self.__class__.__name__, task)
        ack = self._enqueue(task)
        if ack:
            ctx.log.debug("%s: task %s enqueued",
                          self.__class__.__name__, task)

    def _enqueue(self, task):
        raise NotImplementedError("abstract queue")

    def ack(self, task_id):
        raise NotImplementedError("abstract queue")

    def subscribe(self):
        raise NotImplementedError("abstract queue")

    def list_active_channels(self):
        raise NotImplementedError("abstract queue")

    @property
    def tasks(self):
        raise NotImplementedError("abstract queue")



import functools
import re

from collections import namedtuple
from math import ceil
from flask import request, json, make_response

from . import ctx

ArithmeticExpression = namedtuple(
    '_ArithmeticExpression', field_names=["op", "value"])

DEFAULT_DOCUMENTS_PER_PAGE = 10
ARITHMETIC_OPS = (
    "eq",
    "lt",
    "gt",
    "lte",
    "gte",
    "ne",
)
INTEGER_EXPRESSION = re.compile(r"^\d+$")


def get_page(nopaging=False):
    if nopaging:
        return None
    if '_page' in request.values:
        page = request.values['_page']
    else:
        page = 1
    try:
        page = int(page)
    except (TypeError, ValueError):
        page = 1
    return page


def get_limit(nopaging=False):
    if nopaging:
        return None
    default_limit = ctx.cfg.get(
        'documents_per_page', DEFAULT_DOCUMENTS_PER_PAGE)
    if '_limit' in request.values:
        limit = request.values['_limit']
        try:
            limit = int(limit)
        except ValueError:
            limit = default_limit
    else:
        limit = default_limit
    return limit


def get_request_fields(default_fields=None):
    if "_fields" in request.values:
        return request.values["_fields"].split(",")
    return default_fields


def boolean(value):
    return str(value).lower() in ("yes", "true", "1")


def get_boolean_request_param(param_name):
    return boolean(request.values.get(param_name))


def paginated(data, page=None, limit=None, extra=None, transform=None):

    nopaging = get_boolean_request_param("_nopaging")
    if page is None:
        page = get_page(nopaging)
    if limit is None:
        limit = get_limit()

    if isinstance(data, list):
        count = len(data)
        if limit is not None and page is not None:
            data = data[(page - 1) * limit:page * limit]
    elif hasattr(data, "count"):
        count = data.count()
        if limit is not None and page is not None:
            data = data.skip((page-1)*limit).limit(limit)
    else:
        raise RuntimeError(
            "paginated() accepts either cursor objects or lists")

    if transform is not None:
        data = [transform(x) for x in data]

    total_pages = ceil(count / limit) if limit is not None else None

    result = {
        "page": page,
        "total_pages": total_pages,
        "count": count,
        "data": data
    }

    if extra is not None and hasattr(extra, "items"):
        for k, v in extra.items():
            if k not in result:
                result[k] = v

    return result


def default_transform(fields=None):
    def transform(x):
        return x.to_dict(fields=get_request_fields(fields))
    return transform


def filter_expr(flt):
    if flt.find(" ") >= 0:
        tokens = flt.split()
        return {"$in": tokens}
    try:
        return re.compile(flt, re.IGNORECASE)
    # re.compile's can throw multiple different exceptions. We do not care what went wrong
    except Exception:  # pylint: disable=broad-except
        return ""


def arithmetic_expr(expr, try_int=True):
    tokens = expr.split(":")
    expr_op = "eq"
    if len(tokens) == 2 and tokens[0] in ARITHMETIC_OPS:
        expr_op = tokens[0]
        expr = tokens[1]

    if try_int and INTEGER_EXPRESSION.match(expr):
        expr = int(expr)

    return ArithmeticExpression(op=expr_op, value=expr)


def json_body_required(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        from .errors import InputDataError
        if request.json is None:
            raise InputDataError("json data is missing")
        return func(*args, **kwargs)
    return wrapper


def json_response(data, code=200):
    json_kwargs = {}
    if ctx.cfg.get("debug"):
        json_kwargs["indent"] = 4
    return make_response(json.dumps(data, **json_kwargs), code, {'Content-Type': 'application/json'})



from setuptools import setup, find_packages


setup(
    name="glasskit",
    version="3.9.27",
    description="a micro webframework based on flask and pymongo",
    url="https://gitlab.com/viert/glasskit",
    author="Pavel Vorobyov",
    author_email="aquavitale@yandex.ru",
    license="MIT",
    packages=[pkg for pkg in find_packages() if pkg.startswith("glasskit")],
    include_package_data=True,
    package_data={"glasskit": ["*.txt", "*.py", "*.tmpl"]},
    install_requires=[
        "jinja2",
        "flask",
        "pymongo",
        "mongomock",
        "cachelib",
        "lazy_object_proxy",
        "ipython",
        "requests",
        "mtprof",
    ],
    entry_points={"console_scripts": ["glasskit=glasskit.__main__:main",]},
)



from unittest.mock import patch
from ..uorm.models.storable_model import StorableModel
from ..uorm.models.fields import StringField, Field
from .mongo_mock_test import MongoMockTest

CALLABLE_DEFAULT_VALUE = 4


def callable_default():
    return CALLABLE_DEFAULT_VALUE


class TestModel(StorableModel):
    field1: StringField(rejected=True, index=True, default="default_value")
    field2: Field(required=True)
    field3: Field(required=True, default="required_default_value")
    callable_default_field: Field(default=callable_default)


class TestStorableModel(MongoMockTest):

    def setUp(self) -> None:
        super().setUp()
        TestModel.destroy_all()

    def tearDown(self) -> None:
        TestModel.destroy_all()
        super().tearDown()

    def test_defaults(self):
        model = TestModel()
        self.assertEqual(model.field1, "default_value")
        self.assertEqual(model.field3, "required_default_value")
        self.assertEqual(model.callable_default_field, CALLABLE_DEFAULT_VALUE)

    def test_eq(self):
        model = TestModel({"field2": "mymodel"})
        model.save()
        model2 = TestModel.find_one({"field2": "mymodel"})
        self.assertEqual(model, model2)

    def test_reject_on_update(self):
        model = TestModel.create(field1="original_value",
                                 field2="mymodel_reject_test")
        model.save()
        id_ = model._id
        model.update({"field1": "new_value"})
        model = TestModel.find_one({"_id": id_})
        self.assertEqual(model.field1, "original_value")

    def test_update(self):
        model = TestModel.create(field1="original_value",
                                 field2="mymodel_update_test")
        model.save()
        id_ = model._id
        model.update({"field2": "mymodel_updated"})
        model = TestModel.find_one({"_id": id_})
        self.assertEqual(model.field2, "mymodel_updated")

    def test_update_many(self):
        model1 = TestModel.create(field1="original_value",
                                  field2="mymodel_update_test")
        model1.save()
        model2 = TestModel.create(field1="original_value",
                                  field2="mymodel_update_test")
        model2.save()
        model3 = TestModel.create(field1="do_not_modify",
                                  field2="mymodel_update_test")
        model3.save()

        TestModel.update_many({"field1": "original_value"}, {
                              "$set": {"field2": "mymodel_updated"}})
        model1.reload()
        model2.reload()
        model3.reload()

        self.assertEqual(model1.field2, "mymodel_updated")
        self.assertEqual(model2.field2, "mymodel_updated")
        self.assertEqual(model3.field2, "mymodel_update_test")

    @patch("glasskit.uorm.models.storable_model.db.l2c", spec=True)
    @patch("glasskit.uorm.models.storable_model.db.l1c", spec=True)
    def test_invalidate(self, l1c, l2c):

        class Model(StorableModel):

            field1: StringField()

            KEY_FIELD = "field1"
            USE_INITIAL_STATE = True

            def setup_initial_state(self):
                return {
                    "_id": self._id,
                    "field1": self.field1
                }

        model = Model({"field1": "value"})
        model.save()

        l1c.delete.assert_called_once_with("model.value")
        l2c.delete.assert_called_once_with("model.value")

        model.save()
        l1c.delete.assert_any_call(f"model.{model._id}")
        l1c.delete.assert_any_call("model.value")
        l2c.delete.assert_any_call(f"model.{model._id}")
        l2c.delete.assert_any_call("model.value")
